#!/usr/bin/python
# -*- coding:utf-8 -*-

import json
import requests
import logging
import hmac
import hashlib
import base64
import time
from uuid import uuid1
import urllib.parse

try:
    from kraken.exceptions.exceptions import KrakenExceptions
except:
    print('USING LOCAL MODULE')
    sys.path.append('/Users/benjamin/repositories/Trading/python-kraken-sdk')
    from kraken.exceptions.exceptions import KrakenExceptions


class KrakenBaseRestAPI(object):
    ''' Base class for all Spot clients

        Handles un/signed requests and returns exception handled results


        ====== P A R A M E T E R S ======
        key: str, defualt: ''
            Spot API public key
        secret: str, default: ''
            Spot API secret key
        url: str, default: 'https://api.kraken.com'
            optional url
        sandbox: bool, default: False
            not used so far
    '''

    URL = 'https://api.kraken.com'
    API_V = '/0'

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False, **kwargs):
        if sandbox: raise ValueError('Sandbox not availabel for Kraken Spot trading.')
        if url != '': self.url = url
        else: self.url = self.URL

        self.__key = key
        self.__secret = secret

    def _request(self, 
        method: str, 
        uri: str, 
        timeout: int=10, 
        auth: bool=True,
        params: dict={}, 
        do_json: bool=False, 
        return_raw: bool=False
    ) -> dict:
        method = method.upper()
        data_json = ''
        if method in ['GET', 'DELETE']:
            if params:
                strl = []
                for key in sorted(params): strl.append(f'{key}={params[key]}')
                data_json += '&'.join(strl)
                uri += f'?{data_json}'.replace(' ', '%20')

        headers = { 'User-Agent': 'python-kraken-sdk' }
        if auth:
            if not self.__key or self.__key == '' or not self.__secret or self.__secret == '': raise ValueError('Missing credentials')
            params['nonce'] = str(int(time.time() * 1000))
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
                'API-Key': self.__key,
                'API-Sign': self.get_kraken_signature(f'{self.API_V}{uri}', params)
            }

        url = f'{self.url}{self.API_V}{uri}'
        if method in ['GET', 'DELETE']:
            return self.__check_response_data(requests.request(method=method, url=url, headers=headers, timeout=timeout), return_raw)
        elif do_json:
            return self.__check_response_data(requests.request(method=method, url=url, headers=headers, json=params, timeout=timeout), return_raw)
        else:
            return self.__check_response_data(requests.request(method=method, url=url, headers=headers, data=params, timeout=timeout), return_raw)

    def get_kraken_signature(self, urlpath: str, data: dict) -> str:
        return base64.b64encode(
            hmac.new(
                base64.b64decode(self.__secret), 
                urlpath.encode() + hashlib.sha256((str(data['nonce']) + urllib.parse.urlencode(data)).encode()).digest(),
                hashlib.sha512
            ).digest()
        ).decode()

    def __check_response_data(self, response_data, return_raw: bool=False) -> dict:
        if response_data.status_code in [ '200', 200 ]:
            if return_raw: return response_data
            try:
                data = response_data.json()
            except ValueError:
                raise Exception(response_data.content)
            else:
                if 'error' in data:
                    if len(data.get('error')) == 0 and 'result' in data: return data['result']
                    elif data['error'] == 'authenticationError' or 'authenticationError' in data['error']:
                        raise KrakenExceptions.KrakenAuthenticationError(data)
                    elif data['error'] == 'EGeneral:Permission denied' or 'EGeneral:Permission denied' in data['error']:
                        raise KrakenExceptions.KrakenPermissionDeniedError(data)
                    else: raise Exception(f'{response_data.status_code} - {data}')
                else: return data
        else: raise Exception(f'{response_data.status_code}-{response_data.text}')

    @property
    def return_unique_id(self) -> str:
        return ''.join([each for each in str(uuid1()).split('-')])

    def _to_str_list(self, a) -> str:
        if type(a) == str: return a
        elif type(a) == list: return ','.join([i for i in a])
        else: raise ValueError('a must be string or list of strings')

class KrakenBaseFuturesAPI(object):
    ''' Base class for all Futures clients

        Handles un/signed requests and returns exception handled results

        ====== P A R A M E T E R S ======
        key: str, defualt: ''
            Futures API public key
        secret: str, default: ''
            Futures API secret key
        url: str, default: 'https://futures.kraken.com'
            optional url
        sandbox: bool, default: False
            if set to true the url will be 'https://demo-futures.kraken.com'
        
        ====== N O T E S ======
        If the sandbox environment is chosen, the keys must be generated here:
            https://demo-futures.kraken.com/settings/api
    '''

    URL = 'https://futures.kraken.com'
    SANDBOX_URL = 'https://demo-futures.kraken.com'

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False, **kwargs):
        
        self.sandbox = sandbox
        if url: self.url = url
        elif self.sandbox: self.url = self.SANDBOX_URL
        else: self.url = self.URL
        
        self.__key = key
        self.__secret = secret
        self.nonce = 0

    def _request(self, 
        method: str, 
        uri: str, 
        timeout: int=10, 
        auth: bool=True, 
        postParams: dict={},
        queryParams: dict={}, 
        return_raw: bool=False
    ) -> dict:
        method = method.upper()

        postString: str = ''
        if postParams:
            strl: [str] = []
            for key in sorted(postParams): strl.append(f'{key}={postParams[key]}')
            postString = '&'.join(strl)

        queryString: str = ''
        if queryParams:
            strl: [str] = []
            for key in sorted(queryParams): strl.append(f'{key}={queryParams[key]}')
            queryString = '&'.join(strl).replace(' ', '%20')

        headers = { 'User-Agent': 'python-kraken-sdk' }
        if auth:
            if not self.__key or self.__key == '' or not self.__secret or self.__secret == '': raise ValueError('Missing credentials')
            self.nonce = (self.nonce + 1) % 1
            nonce = str(int(time.time() * 1000)) + str(self.nonce).zfill(4)
            headers = {
                'Content-Type': 'application/x-www-form-urlencoded; charset=utf-8',
                'Nonce': nonce,
                'APIKey': self.__key,
                'Authent': self.get_kraken_futures_signature(uri, queryString + postString, nonce)
            }

        if method in ['GET', 'DELETE']:
            return self.__check_response_data(
                requests.request(
                    method=method,
                    url=f'{self.url}{uri}' if queryString == '' else f'{self.url}{uri}?{queryString}', 
                    headers=headers, 
                    timeout=timeout
                ), 
                return_raw
            )
        elif method == 'PUT':
            return self.__check_response_data(
                requests.request(
                method=method, 
                url=f'{self.url}{uri}', 
                params=str.encode(queryString), 
                headers=headers, 
                timeout=timeout
                ), 
                return_raw
            )
        else:
            return self.__check_response_data(
                requests.request(
                    method=method, 
                    url=f'{self.url}{uri}?{postString}', 
                    data=str.encode(postString), 
                    headers=headers, 
                    timeout=timeout
                ), return_raw
            )

    def get_kraken_futures_signature(self, endpoint: str, data: str, nonce: str) -> str:
        # https://github.com/CryptoFacilities/REST-v3-Python/blob/ee89b9b324335d5246e2f3da6b52485eb8391d50/cfRestApiV3.py#L295-L296
        if endpoint.startswith('/derivatives'): endpoint = endpoint[len('/derivatives'):] 
        sha256_hash = hashlib.sha256()
        sha256_hash.update((data + nonce + endpoint).encode('utf8'))
        return base64.b64encode(
            hmac.new(
                base64.b64decode(self.__secret), 
                sha256_hash.digest(), 
                hashlib.sha512
            ).digest()
        )
   
    def __check_response_data(self, response_data, return_raw: bool=False) -> dict:
        if response_data.status_code in [ '200', 200 ]:
            if return_raw: return response_data
            try:
                data = response_data.json()
            except ValueError:
                raise Exception(response_data.content)
            else:
                if 'error' in data:
                    if len(data.get('error')) == 0 and 'result' in data: return data['result']
                    elif data['error'] == 'authenticationError':
                        raise KrakenExceptions.KrakenAuthenticationError(data)
                    elif data['error'] == 'EGeneral:Permission denied' or 'EGeneral:Permission denied' in data['error']:
                        raise KrakenExceptions.KrakenPermissionDeniedError(data)
                    else: raise Exception(f'{response_data.status_code} - {data}')
                else: return data
        else: raise Exception(f'{response_data.status_code}-{response_data.text}')


from kraken.base_api.base_api import KrakenBaseFuturesAPI

class MarketClient(KrakenBaseFuturesAPI):

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False) -> None:
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

    def get_ohlc(self, tick_type: str, symbol: str, resolution: int, from_: int=None, to: int=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-charts-ohlc-get-ohlc
            https://support.kraken.com/hc/en-us/articles/4403284627220-OHLC
        '''
        ttypes = ['spot', 'mark', 'trade']
        resoulutions = ['1m', '5m', '15m', '30m', '1h', '4h', '12h', '1d', '1w']
        if tick_type and tick_type not in ttypes:
            raise ValueError(f'tick_type must be in {ttypes}')
        elif resolution != None and resolution not in resoulutions:
            raise ValueError(f'resolution must be in {resolutions}')

        params = { }
        if from_ != None: params['from'] = from_
        if to != None: params['to'] = to
        return self._request('GET', f'/api/charts/v1/{tick_type}/{symbol}/{resolution}', queryParams=params, auth=False)

    def get_tick_types(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-charts-ohlc-get-tick-types'''
        return self._request('GET', '/api/charts/v1/', auth=False)

    def get_tradeable_products(self, tick_type: str) -> dict:
        '''https://docs.futures.kraken.com/#http-api-charts-ohlc-get-tradeable-products'''
        return self._request('GET', f'/api/charts/v1/{tick_type}', auth=False)

    def get_resolutions(self, tick_type: str, tradeable: str) -> dict:
        '''https://docs.futures.kraken.com/#http-api-charts-ohlc-get-resolutions'''
        return self._request('GET', f'/api/charts/v1/{tick_type}/{tradeable}', auth=False)
    
    def get_fee_schedules(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-fee-schedules-get-fee-schedules
            https://support.kraken.com/hc/en-us/articles/360049269572-Fee-Schedules
        '''
        return self._request('GET', '/derivatives/api/v3/feeschedules', auth=False)

    def get_fee_schedules_vol(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-fee-schedules-get-fee-schedule-volumes'''
        return self._request('GET', '/derivatives/api/v3/feeschedules/volumes', auth=True)
        
    def get_orderbook(self, symbol: str=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-market-data-get-orderbook
            https://support.kraken.com/hc/en-us/articles/360022839551-Order-Book
        '''
        params = {}
        if symbol != None: params['symbol'] = symbol
        return self._request('GET', '/derivatives/api/v3/orderbook', queryParams=params, auth=False)

    def get_tickers(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-market-data-get-tickers
            https://support.kraken.com/hc/en-us/articles/360022839531-Tickers
        '''
        return self._request('GET', '/derivatives/api/v3/tickers', auth=False)

    def get_instruments(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-instrument-details-get-instruments
            https://support.kraken.com/hc/en-us/articles/360022635672-Instruments
        '''
        return self._request('GET', '/derivatives/api/v3/instruments', auth=False)
    
    def get_instruments_status(self, instrument: str=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-instrument-details-get-instrument-status-list
           https://docs.futures.kraken.com/#http-api-trading-v3-api-instrument-details-get-instrument-status
        '''
        if instrument:
            return self._request('GET', f'/derivatives/api/v3/instruments/{instrument}/status', auth=False)
        else:
            return self._request('GET', '/derivatives/api/v3/instruments/status', auth=False)

    def get_trade_history(self, symbol: str=None, lastTime: str=None, **kwargs) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-market-data-get-trade-history
            https://support.kraken.com/hc/en-us/articles/360022839511-History
        '''
        params = {}
        if symbol != None: params['symbol'] = symbol 
        if lastTime != None: params['lastTime'] = lastTime
        params.update(kwargs)
        # if params == {}: raise ValueError('Either symbol or lastTime must be specified!')
        return self._request('GET', '/derivatives/api/v3/history', queryParams=params, auth=False)

    def get_historical_funding_rates(self, symbol: str) -> dict:
        '''https://support.kraken.com/hc/en-us/articles/360061979852-Historical-Funding-Rates'''
        return self._request('GET', '/derivatives/api/v4/historicalfundingrates', queryParams={ 'symbol': symbol }, auth=False)

    def get_leverage_preference(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-multi-collateral-get-the-leverage-setting-for-a-market'''
        return self._request('GET', f'/derivatives/api/v3/leveragepreferences', auth=True)

    def set_leverage_preference(self, symbol: str, maxLeverage: float=None) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-multi-collateral-set-the-leverage-setting-for-a-market'''
        params = { 'symbol': symbol }
        if maxLeverage != None: params['maxLeverage'] = maxLeverage
        return self._request('PUT', '/derivatives/api/v3/leveragepreferences', queryParams=params, auth=True)

    def get_pnl_preference(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-multi-collateral-get-pnl-currency-preference-for-a-market'''
        return self._request('GET', '/derivatives/api/v3/pnlpreferences', auth=True)

    def set_pnl_preference(self, symbol: str, pnlPreference: str) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-multi-collateral-set-pnl-currency-preference-for-a-market'''

        return self._request('PUT', '/derivatives/api/v3/pnlpreferences', queryParams={
            'symbol': symbol,
            'pnlPreference': pnlPreference
        }, auth=True)

    def _get_historical_events(self,
        endpoint: str,
        before: int=None,
        continuation_token: str=None,
        since: int=None,
        sort: str=None,
        tradeable: str=None,
        auth: bool=True,
        **kwargs
    ) -> dict:  
        params = {}
        if before != None: params['before'] = before
        if continuation_token != None: params['continuation_token'] = continuation_token
        if since != None: params['since'] = since
        if sort != None: params['sort'] = sort
        if tradeable != None: params['tradeable'] = tradeable
        params.update(kwargs)
        return self._request('GET', endpoint, postParams=params, auth=auth)

    def get_execution_events(self,
        before: int=None,
        continuation_token: str=None,
        since: int=None,
        sort: str=None,
        tradeable: str=None
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-market-history-get-execution-events'''

        return self._get_historical_events(
            endpoint='/api/history/v2/executions',
            before=before,
            continuation_token=continuation_token,
            since=since,
            sort=sort,
            tradeable=tradeable,
            auth=True
        )

    def get_public_execution_events(self,
        tradeable: str,
        before: int=None,
        continuation_token: str=None,
        since: int=None,
        sort: str=None,
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-market-history-get-public-execution-events
        https://support.kraken.com/hc/en-us/articles/4401755685268-Market-History-Executions
        '''
        return self._get_historical_events(
            endpoint=f'/api/history/v2/market/{tradeable}/executions',
            before=before,
            continuation_token=continuation_token,
            since=since,
            sort=sort,
            auth=False
        )

    def get_public_order_events(self,
        tradeable: str,
        before: int=None,
        continuation_token: str=None,
        since: int=None,
        sort: str=None,
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-market-history-get-public-order-events
        https://support.kraken.com/hc/en-us/articles/4401755906452-Market-History-Orders
        '''
        return self._get_historical_events(
            endpoint=f'/api/history/v2/market/{tradeable}/orders',
            before=before,
            continuation_token=continuation_token,
            since=since,
            sort=sort,
            auth=False
        )

    def get_public_mark_price_events(self,
        tradeable: str,
        before: int=None,
        continuation_token: str=None,
        since: int=None,
        sort: str=None,
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-market-history-get-public-mark-price-events
        https://support.kraken.com/hc/en-us/articles/4401748276116-Market-History-Mark-Price
        '''
        return self._get_historical_events(
            endpoint=f'/api/history/v2/market/{tradeable}/price',
            before=before,
            continuation_token=continuation_token,
            since=since,
            sort=sort,
            auth=False
        )

    def get_order_events(self,
        before: int=None,
        continuation_token: str=None,
        since: int=None,
        sort: str=None,
        tradeable: str=None
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-market-history-get-order-events'''
        return self._get_historical_events(
            endpoint='/api/history/v2/orders',
            before=before,
            continuation_token=continuation_token,
            since=since,
            sort=sort,
            tradeable=tradeable,
            auth=True
        )

    def get_trigger_events(self,
        before: int=None,
        continuation_token: str=None,
        since: int=None,
        sort: str=None,
        tradeable: str=None
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-market-history-get-trigger-events'''
        return self._get_historical_events(
            endpoint='/api/history/v2/triggers',
            before=before,
            continuation_token=continuation_token,
            since=since,
            sort=sort,
            tradeable=tradeable,
            auth=True
        )



from kraken.base_api.base_api import KrakenBaseFuturesAPI

class UserClient(KrakenBaseFuturesAPI):

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False) -> None:
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

    def get_wallets(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-account-information-get-wallets'''
        return self._request('GET', f'/derivatives/api/v3/accounts', auth=True)

    def get_open_orders(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-account-information-get-open-orders'''
        return self._request('GET', f'/derivatives/api/v3/openorders', auth=True)
    
    def get_open_positions(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-account-information-get-open-positions'''
        return self._request('GET', f'/derivatives/api/v3/openpositions', auth=True)

    def get_subaccounts(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-account-information-get-subaccounts'''
        return self._request('GET', f'/derivatives/api/v3/subaccounts', auth=True)

    def get_unwindqueue(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-account-information-get-position-percentile-of-unwind-queue'''
        return self._request('GET', f'/derivatives/api/v3/unwindqueue', auth=True)

    def get_notificatios(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-general-get-notifications'''
        return self._request('GET', '/derivatives/api/v3/notifications', auth=True)

    def get_account_log(self,
        before: int=None,
        count: str=None,
        from_: str=None,
        info: str=None,
        since: str=None,
        sort: str=None,
        to: str=None
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-account-log'''
        params = {}
        if before != None: params['before'] = before
        if count != None: params['count'] = count
        if from_ != None: params['from'] = from_
        if info != None: params['info'] = info
        if since != None: params['since'] = since
        if sort != None: params['sort'] = sort
        if to != None: params['to'] = to
        return self._request('GET', '/api/history/v2/account-log', queryParams=params, auth=True)

    def get_account_log_csv(self) -> dict:
        '''https://docs.futures.kraken.com/#http-api-history-account-log-get-recent-account-log-csv'''
        return self._request('GET', '/api/history/v2/accountlogcsv', auth=True, return_raw=True)





from kraken.base_api.base_api import KrakenBaseRestAPI

class TradeClient(KrakenBaseRestAPI):

    def create_order(
        self,
        ordertype: str,
        side: str,
        volume: str,
        pair: str,
        price: str=None,
        price2: str=None,
        trigger: str=None,
        leverage: str=None,
        stp_type: str='cancel-newest',
        oflags: [str]=None,
        timeinforce: str=None,
        starttm: str='0',
        expiretm: str='0',
        close_ordertype: str=None,
        close_price: str=None,
        close_price2: str=None,
        deadline: str=None,
        validate: bool=False,
        userref: int=None
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/addOrder'''
        params = {
            'ordertype': str(ordertype),
            'type': str(side),
            'volume': str(volume),
            'pair': str(pair),
            'stp_type': stp_type,
            'starttm': starttm,
            'expiretm': expiretm,
            'validate': validate
        }
        if trigger != None:
            if ordertype in ['stop-loss', 'stop-loss-limit', 'take-profit-limit', 'take-profit-limit']:
                if timeinforce != None: params['trigger'] = trigger
                else: raise ValueError(f'Cannot use trigger {trigger} and timeinforce {timeinforce} together')
            else: raise ValueError(f'Cannot use trigger on ordertype {ordertype}')
        elif timeinforce != None: params['timeinforce'] = timeinforce
        if price != None: params['price'] = str(price)
        if price2 != None: params['price2'] = str(price2)
        if leverage != None: params['leverage'] = str(leverage)
        if oflags != None: params['oflags'] = self._to_str_list(oflags)
        if close_ordertype != None: params['close[ordertype]'] = close_ordertype
        if close_price != None: params['close[price]'] = close_price
        if close_price2 != None: params['close[price2]'] = close_price2
        return self._request('POST', '/private/AddOrder', params=params)


    def create_order_batch(self, orders: [dict], pair: str, deadline: str=None, validate: bool=False) -> dict:
        '''https://docs.kraken.com/rest/#operation/addOrderBatch'''
        params = {
            'orders': orders,
            'pair': pair,
            'validate': validate
        }
        if deadline != None: params['deadline'] = deadline
        return self._request('POST', '/private/AddOrderBatch', params=params, do_json=True)

    def edit_order(self, 
        txid: str, 
        pair: str, 
        volume: str=None, 
        price: str=None, 
        price2: str=None, 
        oflags=None, 
        deadline: str=None, 
        cancel_response: bool=None, 
        validate: str=False, 
        userref: int=None
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/editOrder'''
        params = {
            'txid': txid,
            'pair': pair,
            'validate': validate
        }
        if userref != None: params['userref'] = userref
        if volume != None: params['volume'] = volume
        if price != None: params['price'] = price
        if price2 != None: params['price2'] = price2
        if oflags != None: params['oflags'] = self._to_str_list(oflags)
        if cancel_response != None: params['cancel_response'] = cancel_response
        return self._response('POST', '/private/EditOrder', params=params)

    def cancel_order(self, txid) -> dict:
        '''https://docs.kraken.com/rest/#operation/cancelOrder'''
        return self._request('POST', '/private/CancelOrder', params={ 'txid': txid })

    def cancel_all_orders(self) -> dict:
        '''https://docs.kraken.com/rest/#operation/cancelAllOrders'''
        return self._request('POST', '/private/CancelAll')

    def cancel_all_orders_after_x(self, timeout: int) -> dict:
        '''https://docs.kraken.com/rest/#operation/cancelAllOrdersAfter'''
        return self._request('POST', '/private/CancelAllOrdersAfter', params={ 'timeout': timeout })

    def cancel_order_batch(self, orders: [str]) -> dict:
        '''https://docs.kraken.com/rest/#operation/cancelOrderBatch'''
        return self._request('POST', '/private/CancelOrderBatch', params={ 'orders': orders })





from kraken.futures.market.market import MarketClient
from kraken.futures.trade.trade import TradeClient
from kraken.futures.user.user import UserClient
from kraken.futures.funding.funding import FundingClient
from kraken.futures.websocket.websocket import KrakenFuturesWSClientCl
from kraken.exceptions.exceptions import KrakenExceptions 
class User(UserClient):
    pass

class Trade(TradeClient):
    pass

class Market(MarketClient):
    pass

class Funding(FundingClient):
    pass

class KrakenFuturesWSClient(KrakenFuturesWSClientCl):
    pass


from kraken.spot.market.market import MarketClient
from kraken.spot.trade.trade import TradeClient
from kraken.spot.user.user import UserClient
from kraken.spot.funding.funding import FundingClient
from kraken.spot.staking.staking import StakingClient
from kraken.spot.websocket.websocket import KrakenSpotWSClientCl

class User(UserClient):
    pass

class Trade(TradeClient):
    pass

class Market(MarketClient):
    pass

class Funding(FundingClient):
    pass

class Staking(StakingClient):
    pass

class KrakenSpotWSClient(KrakenSpotWSClientCl):
    pass



class KrakenExceptions(object):

    class MaxReconnectError(Exception):
        '''Raised when websocket connection has to many reconnects'''
        def __init__(self, message=''):
            super(Exception, self).__init__(message)

    class KrakenAuthenticationError(Exception):
        '''Raised when credentials are invalid/kraken responses "authenticationError"'''
        def __init__(self, message=''):
            super(Exception, self).__init__(message)

    class KrakenPermissionDeniedError(Exception):
        '''Raised when credentials are valid but permissions are restricted'''
        def __init__(self, message=''):
            super(Exception, self).__init__(message)


from kraken.base_api.base_api import KrakenBaseRestAPI

class UserClient(KrakenBaseRestAPI):

    def get_account_balance(self) -> dict:
        '''https://docs.kraken.com/rest/#operation/getAccountBalance'''
        return self._request('POST', '/private/Balance')

    def get_balances(self, currency: str) -> dict:
        '''Returns the balance and available balance of a given currency'''
        
        balance = float(0)   
        currency_found = False
        for symbol, value in self.get_account_balance().items():
            if balance != float(0): break
            elif symbol in [ currency, f'Z{currency}', f'X{currency}' ]: 
                balance = float(value) 
                currency_found = True
        if not currency_found: raise ValueError('Currency not found!') 

        available_balance = balance
        for txid, order in self.get_open_orders()['open'].items():
            if currency in order['descr']['pair'][0:len(currency)]:
                if order['descr']['type'] == 'sell': 
                    available_balance -= float(order['vol'])
            elif currency in order['descr']['pair'][-len(currency):]:
                if order['descr']['type'] == 'buy': 
                    available_balance -= float(order['vol']) * float(order['descr']['price'])
        
        return {
            'currency': currency,
            'balance': balance,
            'available_balance': available_balance
        }

    def get_trade_balance(self, asset: str=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradeBalance'''
        params = {}
        if asset != None: params['asset'] = asset
        return self._request('POST', '/private/TradeBalance', params=params)

    def get_open_orders(self, trades: bool=False, userref: int=None) -> dict:
        '''https://docs.kraken.com/rest/#operation/getOpenOrders'''
        params = { 'trades': trades }
        if userref != None: params['userref'] = userref
        return self._request('POST', '/private/OpenOrders', params=params)

    def get_closed_orders(self, 
        trades: bool=False, 
        userref: int=None, 
        start: int=None, 
        end: int=None, 
        ofs: int=None, 
        closetime: str='both'
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/getClosedOrders'''
        params = {
            'trades': trades,
            'closetime': closetime
        }
        if userref != None: params['userref'] = userref
        if start != None: params['start'] = start
        if end != None: params['end'] = end
        if ofs != None: params['ofs'] = ofs

        return self._request('POST', '/private/ClosedOrders', params=params)

    def get_orders_info(self, txid, trades: bool=False, userref: int=None) -> dict:
        '''https://docs.kraken.com/rest/#tag/User-Data/operation/getOrdersInfo'''
        params = {
            'txid': txid,
            'trades': trades
        }
        if type(txid) == list: params['txid'] = self._to_str_list(txid)
        if userref != None: params['userref'] = userref
        return self._request('POST', '/private/QueryOrders', params=params)

    def get_trades_history(self, 
        type_: str='all', 
        trades: bool=False, 
        start: int=None, 
        end: int=None, 
        ofs: int=None
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradeHistory'''
        params = {
            'type': type_,
            'trades': trades
        }
        if start != None: params['start'] = start
        if end != None: params['end'] = end
        if ofs != None: params['ofs'] = ofs
        return self._request('POST', '/private/TradesHistory', params=params)

    def get_trades_info(self, txid: str, trades: bool=False) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradesInfo'''
        return self._request('POST', '/private/QueryTrades', params={ 
            'trades': trades,
            'txid': self._to_str_list(txid)
        })

    def get_open_positions(self, txid=None, docalcs: bool=False, consolidation: str='market') -> dict:
        '''https://docs.kraken.com/rest/#operation/getOpenPositions'''
        params = {
            'docalcs': docalcs,
            'consolidation': consolidation
        }
        if txid != None: params['txid'] = self._to_str_list(txid)
        return self._request('POST', '/private/OpenPositions', params=params)

    def get_ledgers_info(self, 
        asset: str='all', 
        aclass: str='currency', 
        type_: str='all', 
        start: int=None, 
        end: int=None, 
        ofs: int=None
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/getLedgers'''
        params = {
            'asset': asset,
            'aclass': aclass,
            'type': type_
        }
        if type(params['asset']) == list: params['asset'] = self._to_str_list(asset)
        if start != None: params['start'] = start
        if end != None: params['end'] = end
        if ofs != None: params['ofs'] = ofs
        return self._request('POST', '/private/Ledgers', params=params)

    def get_ledgers(self, id, trades: bool=False) -> dict:
        '''https://docs.kraken.com/rest/#operation/getLedgersInfo'''
        return self._request('POST', '/private/QueryLedgers', params={ 
            'trades': trades,
            'id': self._to_str_list(id)
        })

    def get_trade_volume(self, pair=None, fee_info: bool=True) -> dict:
        '''https://docs.kraken.com/rest/#operation/getTradeVolume'''
        params = { 'fee-info': fee_info}
        if pair != None: params['pair'] = self._to_str_list(pair)
        return self._request('POST', '/private/TradeVolume', params=params)

    def request_export_report(self, 
        report: str, 
        description: str, 
        format_: str='CSV', 
        fields: str='all', 
        starttm: int=None, 
        endtm: int=None,
        **kwargs
    ) -> dict:
        '''https://docs.kraken.com/rest/#operation/addExport

        ---- RESPONSE ----
            {'id': 'INSG'}
        '''
        if report not in ['trades', 'ledgers']: 
            raise ValueError('report must be one of "trades", "ledgers"')
        params = {
            'report': report,
            'description': description,
            'format': format_,
            'fields': self._to_str_list(fields)
        }
        params.update(kwargs)
        if starttm != None: params['starttm'] = starttm
        if endtm != None: params['endtm'] = endtm
        return self._request('POST', '/private/AddExport', params=params)

    def get_export_report_status(self, report: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/exportStatus
        
        ---- RESPONSE ----
        [{'id': 'INSG', 'descr': 'myLedgers1', 'format': 'CSV', 'report': 'ledgers', 'status': 'Processed', 'aclass': 'currency', 'fields': 'all', 'asset': 'all', 'subtype': 'all', 'starttm': '1656633600', 'endtm': '1657355882', 'createdtm': '1657355882', 'expiretm': '1658565482', 'completedtm': '1657355887', 'datastarttm': '1656633600', 'dataendtm': '1657355882', 'flags': '0'}]
        '''
        if report not in ['trades', 'ledgers']: 
            raise ValueError('report must be one of "trades", "ledgers"')
        return self._request('POST', '/private/ExportStatus', params={ 'report': report })

    def retrieve_export(self, id_: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/retrieveExport'''
        return self._request('POST', '/private/RetrieveExport', params={ 'id': id_ }, return_raw=True)

    def delete_export_report(self, id_: str, type_: str) -> dict:
        '''https://docs.kraken.com/rest/#operation/removeExport'''
        return self._request('POST', '/private/RemoveExport', params={
            'id': id_,
            'type': type_
        })



from kraken.base_api.base_api import KrakenBaseFuturesAPI

class FundingClient(KrakenBaseFuturesAPI):

    def __init__(self, key: str='', secret: str='', url: str='', sandbox: bool=False) -> None:
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

    def get_historical_funding_rates(self, symbol: str) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-historical-funding-rates-historicalfundingrates'''
        return self._request('GET', f'/derivatives/api/v4/historicalfundingrates', queryParams={'symbol': symbol}, auth=False)

    def initiate_wallet_transfer(self, amount: str, fromAccount: str, toAccount: str, unit: str) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-transfers-initiate-wallet-transfer'''
        return self._request('POST', '/derivatives/api/v3/transfer', postParams={
            'amount': amount,
            'fromAccount': fromAccount,
            'toAccount': toAccount,
            'unit': unit
        }, auth=True)

    def initiate_subccount_transfer(self,
        amount: str,
        fromAccount: str, 
        fromUser: str, 
        toAccount: str, 
        toUser: str, 
        unit: str
    ) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-transfers-initiate-sub-account-transfer'''
        return self._request('POST', '/derivatives/api/v3/transfer/subaccount', postParams={
            'amount': amount,
            'fromAccount': fromAccount,
            'fromUser': fromUser,
            'toAccount': toAccount,
            'toUser': toUser,
            'unit': unit
        }, auth=True)

    def initiate_withdrawal_to_spot_wallet(self, amount: str, currency: str, sourceWallet: str=None, **kwargs) -> dict:
        '''https://docs.futures.kraken.com/#http-api-trading-v3-api-transfers-initiate-withdrawal-to-spot-wallet'''
        if self.sandbox: raise ValueError('This function is not available in sandbox mode.')
        params = {
            'amount': str(amount),
            'currency': currency,
            'sourceWallet': sourceWallet
        }
        params.update(kwargs)
        return self._request('POST', '/derivatives/api/v3/withdrawal', postParams=params, auth=True)


import time
import asyncio
import websockets
from random import random
import logging
import traceback
import json
import sys
import copy

try:
    from kraken.futures.ws_client.ws_client import FuturesWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions 
except:
    print('USING LOCAL MODULE')
    sys.path.append('/Users/benjamin/repositories/Trading/python-kraken-sdk')
    from kraken.futures.ws_client.ws_client import FuturesWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions 

class ConnectFuturesWebsocket(object):
    '''
        This class is only called by the KrakenFuturesWSClientCl class
        to establish and handle a websocket connection.

        ====== P A R A M E T E R S ======
        client: kraken.futures.client.KrakenFuturesWSClient
            the websocket client
        endpoint: str
            endpoint/url to connect with
        callback: function [optional], default=None
            callback function to call when a message is received
    '''

    MAX_RECONNECT_NUM = 2

    def __init__(self, client, endpoint: str, callback):
        self.__client = client
        self.__ws_endpoint = endpoint
        self.__callback = callback

        self.__reconnect_num = 0

        self.__last_challenge = None
        self.__new_challenge = None
        self.__challenge_ready = False

        self.__socket = None
        self.__subscriptions = []        
        
        self.__task = asyncio.ensure_future(
            self.__run_forever(), 
            loop=asyncio.get_running_loop()      
        )
    
    @property
    def subscriptions(self) -> list:
        return self.__subscriptions

    async def __run(self, event: asyncio.Event):
        keep_alive = True
        self.__new_challenge = None
        self.__last_challenge = None

        async with websockets.connect(f'wss://{self.__ws_endpoint}', ping_interval=30) as socket:
            logging.info(f'Websocket connected!')            
            self.__socket = socket

            if not event.is_set(): event.set()
            self.__reconnect_num = 0

            while keep_alive:
                try:
                    _msg = await asyncio.wait_for(self.__socket.recv(), timeout=15)
                except asyncio.TimeoutError:
                    pass
                except asyncio.CancelledError:
                    logging.exception('asyncio.CancelledError')
                    keep_alive = False    
                    await self.__callback({'event': 'asyncio.CancelledError'}) 
                else:
                    try:
                        msg = json.loads(_msg)
                    except ValueError:
                        logger.warning(_msg)
                    else:
                        forward = True
                        if 'event' in msg:
                            e = msg['event']
                            if e == 'challenge' and 'message' in msg:
                                forward = False
                                self.__handle_new_challenge(msg)                                
                            elif e == 'subscribed': self.__append_subscription(msg)
                            elif e == 'unsubscribed': self.__remove_subscription(msg)
                        if forward: await self.__callback(msg)

    async def __run_forever(self) -> None:
        try:
            while True: await self.__reconnect()
        except KrakenExceptions.MaxReconnectError: 
            await self.__callback({'event': 'KrakenExceptions.MaxReconnectError'})
        except Exception as e:
            # for task in asyncio.all_tasks(): task.cancel()
            logging.error(traceback.format_exc())
        # except asyncio.CancelledError: pass
        finally: self.__client.exception_occur = True

    async def __reconnect(self):
        logging.info('Websocket start connect/reconnect')

        self.__reconnect_num += 1
        if self.__reconnect_num >= self.MAX_RECONNECT_NUM:
            raise KrakenExceptions.MaxReconnectError()

        reconnect_wait = self.__get_reconnect_wait(self.__reconnect_num)
        logging.debug(f'asyncio sleep reconnect_wait={reconnect_wait} s reconnect_num={self.__reconnect_num}')
        await asyncio.sleep(reconnect_wait)
        logging.debug(f'asyncio sleep done')
        event = asyncio.Event()

        tasks = {
            asyncio.ensure_future(self.__recover_subscription_req_msg(event)): self.__recover_subscription_req_msg,
            asyncio.ensure_future(self.__run(event)): self.__run
        }

        while set(tasks.keys()):
            finished, pending = await asyncio.wait(tasks.keys(), return_when=asyncio.FIRST_EXCEPTION)
            exception_occur = False
            for task in finished:
                if task.exception():
                    exception_occur = True
                    message = f'{task} got an exception {task.exception()}\nTRACEBACK: {traceback.format_exc()}'
                    logging.warning(message)
                    for process in pending:
                        logging.warning(f'pending {process}')
                        try: process.cancel()
                        except asyncio.CancelledError: logging.exception('CancelledError')
                        logging.warning('cancel ok')
                    await self.__callback({ 'ws-error': message })
            if exception_occur: break            
        logging.warning('reconnect over')

    async def __recover_subscription_req_msg(self, event) -> None:
        logging.info(f'Recover subscriptions {self.__subscriptions} waiting.')
        await event.wait()
        
        for sub in self.__subscriptions:            
            if sub['feed'] in self.__client.get_available_private_subscription_feeds():
                await self.send_message(copy.deepcopy(sub), private=True)
            elif sub['feed'] in self.__client.get_available_public_subscription_feeds():
                await self.send_message(copy.deepcopy(sub), private=False)
            else: pass
            logging.info(f'{sub}: OK')

        logging.info(f'Recover subscriptions {self.__subscriptions} done.')

    async def send_message(self, msg: dict, private: bool=False) -> None:
        while not self.__socket: await asyncio.sleep(.4)
        
        if private:
            if not self.__client.isAuth:
                raise ValueError('Cannot access private endpoints with unauthenticated client!')
            elif not self.__challenge_ready: await self.__check_challenge_ready()

            msg['api_key'] = self.__client._key
            msg['original_challenge'] = self.__last_challenge
            msg['signed_challenge'] = self.__new_challenge
            
        await self.__socket.send(json.dumps(msg))

    def __handle_new_challenge(self, msg: dict) -> None:
        self.__last_challenge = msg['message']
        self.__new_challenge = self.__client._get_sign_challenge(self.__last_challenge)
        self.__challenge_ready = True

    async def __check_challenge_ready(self) -> None:
        await self.__socket.send(json.dumps({
            'event': 'challenge',
            'api_key': self.__client._key
        }))

        logging.debug('Awaiting challenge...')
        while not self.__challenge_ready: await asyncio.sleep(.1)

    def __get_reconnect_wait(self, attempts: int) -> float:
        return round(random() * min(60*3, (2 ** attempts) - 1) + 1)

    def __append_subscription(self, msg: dict) -> None:
        self.__remove_subscription(msg=msg) # remove from list, to avoid duplicates
        sub = self.__build_subscription(msg)
        self.__subscriptions.append(sub)

    def __remove_subscription(self, msg: dict) -> None:
        sub = self.__build_subscription(msg)
        self.__subscriptions = [x for x in self.__subscriptions if x != sub]

    def __build_subscription(self, subscription: dict) -> dict:
        sub = { 'event': 'subscribe' }

        if 'event' not in subscription                                        \
            or subscription['event'] not in ['subscribed', 'unsubscribed']    \
            or 'feed' not in subscription: 
            raise ValueError('Cannot append/remove subscription with missing attributes.')
        
        elif subscription['feed'] in self.__client.get_available_public_subscription_feeds(): # public subscribe
            if 'product_ids' in subscription:
                sub['product_ids'] = subscription['product_ids'] if type(subscription['product_ids']) == list else [subscription['product_ids']]
            sub['feed'] = subscription['feed']
        
        elif subscription['feed'] in self.__client.get_available_private_subscription_feeds(): # private subscription
            sub['feed'] = subscription['feed']
        else: logging.warn('Feed not implemented. Please contact the python-kraken-sdk package author.')
        return sub

    def get_active_subscriptions(self) -> [dict]:
        return self.__subscriptions


class KrakenFuturesWSClientCl(FuturesWsClientCl):
    '''https://docs.futures.kraken.com/#websocket-api
    
        Class to access public and (optional) 
        private/authenticated websocket connection.

        ====== P A R A M E T E R S ======
        key: str, [optional], default: ''
            API Key for the Kraken API
        secret: str, [optional], default: ''
            Secret API Key for the Kraken API
        url: str, [optional], default: ''
            Set a specific/custom url
        callback: async function [optional], default=None
            callback function which receives the websocket messages
        sandbox: bool [optional], default=False
            use the Kraken Futures demo url

        ====== E X A M P L E ======
        import asyncio
        from kraken.futures.client import KrakenFuturesWSClient

        async def main() -> None:

            # ___Custom_Trading_Bot__________
            class Bot(KrakenFuturesWSClient):

                async def on_message(self, event) -> None:
                    print(event)

            bot = Bot() # unauthenticated 
            auth_bot = Bot(key=key, secret=secret) # authenticated

            # ... now call for example subscribe and so on

            while True: await asyncio.sleep(6)

        if __name__ == '__main__':
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try: asyncio.run(main())
            except KeyboardInterrupt: loop.close()
    '''

    PROD_ENV_URL = 'futures.kraken.com/ws/v1'
    DEMO_ENV_URL = 'demo-futures.kraken.com/ws/v1'

    def __init__(self, key: str='', secret: str='', url: str='', callback=None, sandbox: bool=False): 
        super().__init__(key=key, secret=secret, url=url, sandbox=sandbox)

        self.exception_occur = False
        self.__callback = callback        
        self._conn = ConnectFuturesWebsocket(
            client=self,
            endpoint= url if url != '' else self.DEMO_ENV_URL if sandbox else self.PROD_ENV_URL,
            callback=self.on_message 
        )

    async def on_message(self, msg) -> None:
        ''' Calls the defined callback function (if defined) 
            or overload this function
            
            ====== P A R A M E T E R S ======
            msg: dict
                message received from Kraken via the websocket connection

            ====== N O T E S ======
            Can be overloaded like in the documentation of this class.
        '''
        if self.__callback != None: await self.__callback(msg)
        else:
            logging.warning('Received event but no callback is defined')
            loggin.info(msg)

    async def subscribe(self, feed: str, products: [str]=None) -> None:
        '''Subscribe to a channel/feed
            https://docs.futures.kraken.com/#websocket-api-websocket-api-introduction-subscriptions

            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to subscribe to
            products: [str]
                list of assets or list of a single product

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            await bot.subscribe(feed='ticker', products=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ====== 
            Success or failures are sent over the websocket connection and can be  
            received via the on_message callback function.
        '''

        message = { 'event': 'subscribe', 'feed': feed }

        if products != None: 
            if type(products) != list: raise ValueError('Parameter products must be type of [str] (e.g. pair=["PI_XBTUSD"])')
            else: message['product_ids'] = products     
            
        if feed in self.get_available_private_subscription_feeds():
            if products != None: raise ValueError('There is no private feed that accepts products!')
            else: await self._conn.send_message(message, private=True)
        elif feed in self.get_available_public_subscription_feeds():
            if products != None:
                for product in products:
                    sub = copy.deepcopy(message)
                    sub['product_ids'] = [product]
                    await self._conn.send_message(sub, private=False)
            else: await self._conn.send_message(message, private=False)
        else: raise ValueError(f'Feed: {feed} not found. Not subscribing to it.')

    async def unsubscribe(self, feed: str, products: [str]=None) -> None:
        '''Unsubscribe from a topic/feed
            https://docs.futures.kraken.com/#websocket-api-websocket-api-introduction-subscriptions
        
            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to unsubscribe from
            products: [str]
                list of assets or list of a single product

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            bot.unsubscribe(feed='ticker', products=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ====== 
            Success or failures are sent over the websocket connection and can be  
            received via the on_message callback function.
        '''

        message = { 'event': 'unsubscribe', 'feed': feed }

        if products != None: 
            if type(products) != list: raise ValueError('Parameter products must be type of [str] (e.g. pair=["PI_XBTUSD"])')
            else: message['product_ids'] = products     

        if feed in self.get_available_private_subscription_feeds():
            if products != None: raise ValueError('There is no private feed that accepts products!')
            else: await self._conn.send_message(message, private=True)
        elif feed in self.get_available_public_subscription_feeds():
            if products != None:
                for product in products:
                    sub = copy.deepcopy(message)
                    sub['product_ids'] = [product]
                    await self._conn.send_message(sub, private=False)                    
            else: await self._conn.send_message(message, private=False)
        else: raise ValueError(f'Feed: {feed} not found. Not unsubscribing it.')

    @staticmethod
    def get_available_public_subscription_feeds() -> [str]:
        return [ 'trade', 'book', 'ticker', 'ticker_lite', 'heartbeat']

    @staticmethod
    def get_available_private_subscription_feeds() -> [str]:
        return [ 
            'fills', 
            'open_positions', 'open_orders', 'open_orders_verbose',
            'balances', 
            'deposits_withdrawals', 'account_balances_and_margins', 
            'account_log', 'notifications_auth'
        ]

    @property
    def isAuth(self) -> bool:
        return self._key and self._secret

    def get_active_subscriptions(self) -> [dict]:
        return self._conn.get_active_subscriptions()


import json
import time
import asyncio
import websockets
from random import random
from uuid import uuid4
import logging
import traceback
import copy

try:
    from kraken.spot.ws_client.ws_client import SpotWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions 
except:
    print('USING LOCAL MODULE')
    sys.path.append('/Users/benjamin/repositories/Trading/python-kraken-sdk')
    from kraken.spot.client.ws_client import SpotWsClientCl
    from kraken.exceptions.exceptions import KrakenExceptions 

class ConnectSpotWebsocket:
    '''
        This class is only called by the KrakenSpotWSClientCl class
        to establish and handle a websocket connection.

        ====== P A R A M E T E R S ======
        client: kraken.spot.client.KrakenSpotWSClient
            the websocket client
        endpoint: str
            endpoint/url to connect with
        callback: function [optional], default=None
            callback function to call when a message is received
        private: bool [optional], default=False
            if client is authenticated to send signed messages
            and get private feeds

    '''

    MAX_RECONNECT_NUM = 10

    def __init__(self, client, endpoint: str, callback, isAuth: bool=False):
        self.__client = client
        self.__ws_endpoint = endpoint
        self.__callback = callback 

        self.__reconnect_num = 0
        self.__ws_conn_details = None

        self.__isAuth = isAuth

        self.__last_ping = None
        self.__socket = None
        self.__subscriptions = []

        self.__task = asyncio.ensure_future(
            self.__run_forever(),
            loop=asyncio.get_running_loop()
        )

    @property
    def subscriptions(self) -> list:
        return self.__subscriptions

    @property
    def isAuth(self) -> bool:
        return self.__isAuth

    async def __run(self, event: asyncio.Event):
        keep_alive = True
        self.__last_ping = time.time()
        self.__ws_conn_details = None if not self.__isAuth else self.__client.get_ws_token()
        logging.debug(f'Websocket token: {self.__ws_conn_details}')

        async with websockets.connect(f'wss://{self.__ws_endpoint}', ping_interval=None) as socket:
            logging.info(f'Websocket connected!')
            self.__socket = socket

            if not event.is_set():
                await self.send_ping()
                event.set()
            self.__reconnect_num = 0

            while keep_alive:
                if time.time() - self.__last_ping > 10: await self.send_ping()
                try:
                    _msg = await asyncio.wait_for(self.__socket.recv(), timeout=15)
                except asyncio.TimeoutError:
                    await self.send_ping()
                except asyncio.CancelledError:
                    logging.exception('asyncio.CancelledError')
                    keep_alive = False
                    await self.__callback({'event': 'asyncio.CancelledError'})
                else:
                    try:
                        msg = json.loads(_msg)
                    except ValueError:
                        logger.warning(_msg)
                    else: 
                        if 'event' in msg:
                            if msg['event'] == 'subscriptionStatus' and 'status' in msg:
                                # remove/assign un/subscriptions
                                try:
                                    if msg['status'] == 'subscribed':
                                        self.__append_subscription(msg)    
                                    elif msg['status'] == 'unsubscribed': 
                                        self.__remove_subscription(msg)  
                                    elif msg['status'] == 'error': logging.warn(msg)
                                except AttributeError: pass
                        await self.__callback(msg)

    async def __run_forever(self) -> None:
        try:
            while True: await self.__reconnect()
        except KrakenExceptions.MaxReconnectError: 
            await self.__callback({'event': 'KrakenExceptions.MaxReconnectError'})
        except Exception as e:
            logging.error(traceback.format_exc())
        finally: self.__client.exception_occur = True

    async def __reconnect(self):
        logging.info('Websocket start connect/reconnect')

        self.__reconnect_num += 1
        if self.__reconnect_num >= self.MAX_RECONNECT_NUM:
            raise KrakenExceptions.MaxReconnectError()

        reconnect_wait = self.__get_reconnect_wait(self.__reconnect_num)
        logging.debug(f'asyncio sleep reconnect_wait={reconnect_wait} s reconnect_num={self.__reconnect_num}')
        await asyncio.sleep(reconnect_wait)
        logging.debug(f'asyncio sleep done')
        event = asyncio.Event()

        tasks = {
            asyncio.ensure_future(self.__recover_subscriptions(event)): self.__recover_subscriptions,
            asyncio.ensure_future(self.__run(event)): self.__run
        }

        while set(tasks.keys()):
            finished, pending = await asyncio.wait(tasks.keys(), return_when=asyncio.FIRST_EXCEPTION)
            exception_occur = False
            for task in finished:
                if task.exception():
                    exception_occur = True
                    message = f'{task} got an exception {task.exception()}\nTRACEBACK: {traceback.format_exc()}'
                    logging.warning(message)
                    for process in pending:
                        logging.warning(f'pending {process}')
                        try: process.cancel()
                        except asyncio.CancelledError: logging.exception('asyncio.CancelledError')
                        logging.warning('Cancel OK')
                    await self.__callback({ 'ws-error': message })
            if exception_occur: break
        logging.warning('reconnect over')

    async def __recover_subscriptions(self, event):
        logging.info(f'Recover {"auth" if self.__isAuth else "public"} subscriptions {self.__subscriptions} waiting.')
        await event.wait()

        for sub in self.__subscriptions:
            cpy = copy.deepcopy(sub)
            private = False
            if 'subscription' in sub                \
                and 'name' in sub['subscription']   \
                and sub['subscription']['name'] in self.__client.private_sub_names:
                cpy['subscription']['token'] = self.__ws_conn_details['token']
                private = True            
            await self.send_message(cpy, private=private)
            logging.info(f'{sub} OK')

        logging.info(f'Recovering {"auth" if self.__isAuth else "public"} subscriptions {self.__subscriptions} done.')

    async def send_ping(self):
        msg = {
            'event': 'ping',
            'reqid': int(time.time() * 1000),
        }
        await self.__socket.send(json.dumps(msg))
        self.__last_ping = time.time()

    async def send_message(self, msg, private: bool=False, retry_count: int=0):
        while not self.__socket: await asyncio.sleep(.4)

        if private and not self.__isAuth:
            raise ValueError('Cannot send private message with public websocket.')
        else:
            msg['reqid'] = int(time.time() * 1000)
            if private and 'subscription' in msg: msg['subscription']['token'] = self.__ws_conn_details['token']
            elif private: msg['token'] = self.__ws_conn_details['token']
            await self.__socket.send(json.dumps(msg))

    def __append_subscription(self, msg: dict) -> None:
        ''' Add a dictionary containing subscription information to list
            This is used to recover when the connection gets interrupted.       

            This function should only be called in 
            when self.__run receives a msg and the following conditions met:
                'event' in msg                           \
                and msg['event'] == 'subscriptionStatus' \
                and 'status' in msg                      \
                and msg['status'] == 'subscribed'

            ====== P A R A M E T E R S ======
            msg: dict
                must be like: { 'subscription': { 'name': '<placeholder>', 'placeholder': 'placeholder', ... }} 
        '''
        self.__remove_subscription(msg)# remove from list, to avoid duplicates
        self.__subscriptions.append(self.__build_subscription(msg))

    def __remove_subscription(self, msg: dict) -> None:
        ''' Remove a dictionary containing subscription information from list.

            This function should only be called in 
            when self.__run receives a msg and the following conditions met:
                'event' in msg                           \
                and msg['event'] == 'subscriptionStatus' \
                and 'status' in msg                      \
                and msg['status'] == 'unsubscribed'

            ====== P A R A M E T E R S ======
            msg: dict
                must be like: { 'subscription': { 'name': '<placeholder>', 'placeholder': 'placeholder', ... }} 
        '''
        sub = self.__build_subscription(msg)
        self.__subscriptions = [x for x in self.__subscriptions if x != sub]

    def __build_subscription(self, msg: dict) -> dict:
        sub = { 'event': 'subscribe' }

        if not 'subscription' in msg or 'name' not in msg['subscription']: 
            raise ValueError('Cannot remove subscription with missing attributes.')
        elif msg['subscription']['name'] in self.__client.public_sub_names: # public endpoint
            if 'pair' in msg: 
                sub['pair'] = msg['pair'] if type(msg['pair']) == list else [msg['pair']]
            sub['subscription'] = msg['subscription']
        elif msg['subscription']['name'] in self.__client.private_sub_names: # private endpoint
            sub['subscription'] = { 'name': msg['subscription']['name'] }
        else: logging.warn('Feed not implemented. Please contact the python-kraken-sdk package author.')
        return sub

    def __get_reconnect_wait(self, attempts: int) -> float:
        return round(random() * min(60*3, (2 ** attempts) - 1) + 1)

class KrakenSpotWSClientCl(SpotWsClientCl):
    '''https://docs.kraken.com/websockets/#overview
    
        Class to access public and (optional) 
        private/authenticated websocket connection.

        This class holds up to two websocket connections, one private
        and one public. 

        ====== P A R A M E T E R S ======
        key: str, [optional], default: ''
            API Key for the Kraken API
        secret: str, [optional], default: ''
            Secret API Key for the Kraken API
        url: str, [optional], default: ''
            Set a specific/custom url
        callback: async function [optional], default=None
            callback function which receives the websocket messages
        beta: bool [optional], default=False
            use the Kraken beta url

        ====== P R O P E R T I E S ======
        public_sub_names: [str]
            list of available public subscription names
        private_sub_names: [str]
            list of available private subscription names
        active_public_subscriptions: [dict]
            list of active public subscriptions
        active_private_subscriptions: [dict]
            list of active private subscriptions

        ====== E X A M P L E ======
        import asyncio
        from kraken.spot.client import KrakenSpotWSClient

        async def main() -> None:
            class Bot(KrakenSpotWSClient):

                async def on_message(self, event) -> None:
                    print(event)
            
            bot = Bot() # unauthenticated
            auth_bot = Bot(key='kraken-api-key', secret='kraken-secret-key') # authenticated
            
            # ... now call for example subscribe and so on

            while True: await asyncio.sleep(6)

        if __name__ == '__main__':
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            try: asyncio.run(main())
            except KeyboardInterrupt: loop.close()
    '''

    PROD_ENV_URL = 'ws.kraken.com'
    AUTH_PROD_ENV_URL = 'ws-auth.kraken.com'
    BETA_ENV_URL = 'beta-ws.kraken.com'
    AUTH_BETA_ENV_URL = 'beta-ws-auth.kraken.com'

    def __init__(self, key: str='', secret: str='', url: str='', callback=None, beta: bool=False):
        super().__init__(key=key, secret=secret, sandbox=beta)
        self.__callback = callback
        self.__isAuth = key and secret        
        
        self.exception_occur = False
        self._pub_conn = ConnectSpotWebsocket(
            client=self,
            endpoint=self.PROD_ENV_URL if not beta else BETA_ENV_URL,
            isAuth=False,
            callback=self.on_message
        )

        self._priv_conn = ConnectSpotWebsocket(
             client=self,
             endpoint=self.AUTH_PROD_ENV_URL if not beta else AUTH_BETA_ENV_URL,
             isAuth=True,
             callback=self.on_message
        ) if self.__isAuth else None

    async def on_message(self, msg: dict):
        ''' Calls the defined callback function (if defined) 
            or overload this function
            
            ====== P A R A M E T E R S ======
            msg: dict
                message received from Kraken via the websocket connection

            ====== N O T E S ======
            Can be overloaded like in the documentation of this class.
        '''
        if self.__callback != None: await self.__callback(msg)
        else:
            logging.warning('Received event but no callback is defined')
            print(msg)

    async def subscribe(self, subscription: dict, pair: [str]=None) -> None:
        '''Subscribe to a channel
            https://docs.kraken.com/websockets-beta/#message-subscribe
            
            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to subscribe to
            pair: [str]
                list of asset pairs or list of a single pair

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            await bot.subscribe(subscription={"name": ticker}, pair=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ====== 
            Success or failures are sent over the websocket connection and can be  
            received via the on_message callback function.
        '''

        if 'name' not in subscription: raise AttributeError('Subscription requires a "name" key."')
        private = True if subscription['name'] in self.private_sub_names else False

        payload = { 
            'event': 'subscribe',
            'subscription': subscription
        }
        if pair != None: 
            if type(pair) != list: raise ValueError('Parameter pair must be type of [str] (e.g. pair=["XBTUSD"])')
            else: payload['pair'] = pair        

        if private: # private == without pair
            if not self.__isAuth: raise ValueError('Cannot subscribe to private feeds without valid credentials!')
            elif pair != None: raise ValueError('Cannot subscribe to private endpoint with specific pair!')
            else: await self._priv_conn.send_message(payload, private=True)

        elif pair != None: # public with pair
            for p in pair:
                sub = copy.deepcopy(payload)
                sub['pair'] = [p]
                await self._pub_conn.send_message(sub, private=False)

        else: await self._pub_conn.send_message(payload, private=False)

    async def unsubscribe(self, subscription: dict, pair: [str]=None) -> None:
        '''Unsubscribe from a topic
            https://docs.kraken.com/websockets/#message-unsubscribe

            ====== P A R A M E T E R S ======
            subscription: dict
                the subscription to unsubscribe from
            pair: [str]
                list of asset pairs or list of a single pair

            ====== E X A M P L E ======
            # ... initialize bot as documented on top of this class.
            bot.unsubscribe(subscription={"name": ticker}, pair=["XBTUSD", "DOT/EUR"])

            ====== N O T E S ====== 
            Success or failures are sent over the websocket connection and can be  
            received via the on_message callback function.
        '''
        if 'name' not in subscription: raise AttributeError('Subscription requires a "name" key."')
        private = True if subscription['name'] in self.private_sub_names else False

        payload = { 
            'event': 'unsubscribe',
            'subscription': subscription
        }
        if pair != None: 
            if type(pair) != list: raise ValueError('Parameter pair must be type of [str] (e.g. pair=["XBTUSD"])')
            else: payload['pair'] = pair
        
        if private: # private == without pair
            if not self.__isAuth: raise ValueError('Cannot unsubscribe from private feeds without valid credentials!')
            elif pair != None: raise ValueError('Cannot unsubscribe from private endpoint with specific pair!')
            else: await self._priv_conn.send_message(payload, private=True)
                
        elif pair != None: # public with pair
            for p in pair:
                sub = copy.deepcopy(payload)
                sub['pair'] = [p]
                await self._pub_conn.send_message(sub, private=False)
                               
        else: await self._pub_conn.send_message(payload, private=False)

    @property
    def private_sub_names(self) -> [str]:
        return  ['ownTrades', 'openOrders']

    @property
    def public_sub_names(self) -> [str]:
        return ['ticker', 'spread', 'book', 'ohlc', 'trade', '*']

    @property
    def active_public_subscriptions(self) -> [dict]:
        if self._pub_conn != None:
            return self._pub_conn.subscriptions 
        else: raise ConnectionError('Public connection does not exist!')

    @property
    def active_private_subscriptions(self) -> [str]:
        if self._priv_conn != None:
            return self._priv_conn.subscriptions
        else: raise ConnectionError('Private connection does not exist!')


